@@ -78,7 +78,7 @@
 	egg_memory_unlock ();
 
 static int lock_warning = 1;
-
+int egg_secure_warnings = 1;
 
 /* 
  * We allocate all memory in units of sizeof(void*). This 
@@ -341,7 +341,7 @@ sec_insert_cell_ring (Cell **ring, Cell *cell)
 	/* Insert back into the mix of available memory */ 
 	if (*ring) { 
 		cell->next = (*ring)->next;
-		cell->prev = (*ring)->prev;
+		cell->prev = *ring;
 		cell->next->prev = cell;
 		cell->prev->next = cell;
 	} else {
@@ -350,6 +350,8 @@ sec_insert_cell_ring (Cell **ring, Cell *cell)
 	}
 	
 	*ring = cell;
+	ASSERT (cell->next->prev == cell);
+	ASSERT (cell->prev->next == cell);
 }
 
 static void
@@ -359,7 +361,10 @@ sec_remove_cell_ring (Cell **ring, Cell *cell)
 	ASSERT (*ring);
 	ASSERT (cell->next);
 	ASSERT (cell->prev);
-	
+
+	ASSERT (cell->next->prev == cell);
+	ASSERT (cell->prev->next == cell);
+
 	if (cell == *ring) {
 		/* The last meta? */
 		if (cell->next == cell) {
@@ -625,7 +630,7 @@ sec_realloc (Block *block, void *memory, size_t length)
 	/* How many words we actually want */
 	n_words = sec_size_to_words (length) + 2;
 
-	/* Less memory is required */
+	/* Less memory is required than is in the cell */
 	if (n_words <= cell->n_words) {
 
 		/* TODO: No shrinking behavior yet */
@@ -636,7 +641,15 @@ sec_realloc (Block *block, void *memory, size_t length)
 		VALGRIND_MAKE_MEM_DEFINED (alloc, length);
 #endif
 		
-		return sec_clear_memory (alloc, valid, length);
+		/* 
+		 * Even though we may be reusing the same cell, that doesn't
+		 * mean that the allocation is shrinking. It could have shrunk
+		 * and is now expanding back some. 
+		 */ 
+		if (length < valid)
+			return sec_clear_memory (alloc, length, valid);
+		else
+			return alloc;
 	}
 	
 	/* Need braaaaaiiiiiinsss... */
@@ -648,7 +661,7 @@ sec_realloc (Block *block, void *memory, size_t length)
 			break;
 		
 		/* Eat the whole neighbor if not too big */
-		if (n_words - cell->n_words >= other->n_words + WASTE) {
+		if (n_words - cell->n_words + WASTE >= other->n_words) {
 			cell->n_words += other->n_words;
 			sec_write_guards (cell);
 			sec_remove_cell_ring (&block->unused, other);
@@ -719,6 +732,45 @@ sec_allocated (Block *block, void *memory)
 	return cell->allocated;
 }
 
+static void
+sec_validate (Block *block)
+{
+	Cell *cell;
+	word_t *word, *last;
+	
+	word = block->words;
+	last = word + block->n_words;
+
+	for (;;) {
+		ASSERT (word < last);
+
+		ASSERT (sec_is_valid_word (block, word));
+		ASSERT (pool_valid (*word));
+		cell = *word;
+	
+		/* Validate that it's actually for real */
+		sec_check_guards (cell);
+	
+		/* Is it an allocated block? */
+		if (cell->allocated > 0) {
+			ASSERT (cell->next == NULL);
+			ASSERT (cell->prev == NULL);
+			ASSERT (cell->allocated <= (cell->n_words - 2) * sizeof (word_t));
+		
+			/* An unused block */
+		} else {
+			ASSERT (cell->next);
+			ASSERT (cell->prev);
+			ASSERT (cell->next->prev == cell);
+			ASSERT (cell->prev->next == cell);
+		}
+		
+		word += cell->n_words;
+		if (word == last)
+			break;
+	}
+}
+
 /* -----------------------------------------------------------------------------
  * LOCKED MEMORY
  */
@@ -739,7 +791,7 @@ sec_acquire_pages (size_t *sz)
 #if defined(HAVE_MLOCK)
 	pages = mmap (0, *sz, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANON, -1, 0);
 	if (pages == MAP_FAILED) {
-		if (lock_warning)
+		if (lock_warning && egg_secure_warnings)
 			fprintf (stderr, "couldn't map %lu bytes of private memory: %s\n", 
 			         (unsigned long)*sz, strerror (errno));
 		lock_warning = 0;
@@ -747,7 +799,7 @@ sec_acquire_pages (size_t *sz)
 	}
 	
 	if (mlock (pages, *sz) < 0) {
-		if (lock_warning && errno != EPERM) {
+		if (lock_warning && egg_secure_warnings && errno != EPERM) {
 			fprintf (stderr, "couldn't lock %lu bytes of private memory: %s\n", 
 			         (unsigned long)*sz, strerror (errno));
 			lock_warning = 0;
@@ -762,7 +814,7 @@ sec_acquire_pages (size_t *sz)
 	return pages;
 	
 #else
-	if (lock_warning)
+	if (lock_warning && egg_secure_warnings)
 		fprintf (stderr, "your system does not support private memory");
 	lock_warning = 0;
 	return NULL;
@@ -777,10 +829,10 @@ sec_release_pages (void *pages, size_t sz)
 	ASSERT (sz % getpagesize () == 0);
 	
 #if defined(HAVE_MLOCK)
-	if (munlock (pages, sz) < 0)
+	if (munlock (pages, sz) < 0 && egg_secure_warnings)
 		fprintf (stderr, "couldn't unlock private memory: %s\n", strerror (errno));
 		
-	if (munmap (pages, sz) < 0)
+	if (munmap (pages, sz) < 0 && egg_secure_warnings)
 		fprintf (stderr, "couldn't unmap private anonymous memory: %s\n", strerror (errno));
 		
 	DEBUG_ALLOC ("gkr-secure-memory: freed block ", sz);
@@ -897,8 +949,9 @@ egg_secure_alloc_full (size_t length, int flags)
 	void *memory = NULL;
 		
 	if (length > 0xFFFFFFFF / 2) {
-		fprintf (stderr, "tried to allocate an insane amount of memory: %lu\n", 
-		         (unsigned long)length);   
+		if (egg_secure_warnings)
+			fprintf (stderr, "tried to allocate an insane amount of memory: %lu\n", 
+			         (unsigned long)length);   
 		return NULL;
 	}
 
@@ -955,9 +1008,9 @@ egg_secure_realloc_full (void *memory, size_t length, int flags)
 	void *alloc = NULL;
 	
 	if (length > 0xFFFFFFFF / 2) {
-		fprintf (stderr, "tried to allocate an insane amount of memory: %lu\n", 
-		         (unsigned long)length);
-		ASSERT (0 && "tried to allocate an insane amount of memory");
+		if (egg_secure_warnings)
+			fprintf (stderr, "tried to allocate an insane amount of memory: %lu\n", 
+			         (unsigned long)length);
 		return NULL;
 	}
 	
@@ -1009,8 +1062,9 @@ egg_secure_realloc_full (void *memory, size_t length, int flags)
 			 */
 			return egg_memory_fallback (memory, length);
 		} else {
-			fprintf (stderr, "memory does not belong to gnome-keyring: 0x%08lx\n", 
-			         (unsigned long)memory);
+			if (egg_secure_warnings)
+				fprintf (stderr, "memory does not belong to gnome-keyring: 0x%08lx\n", 
+				         (unsigned long)memory);
 			ASSERT (0 && "memory does does not belong to gnome-keyring");
 			return NULL;
 		}
@@ -1070,8 +1124,9 @@ egg_secure_free_full (void *memory, int flags)
 		if ((flags & GKR_SECURE_USE_FALLBACK)) {
 			egg_memory_fallback (memory, 0);
 		} else {
-			fprintf (stderr, "memory does not belong to gnome-keyring: 0x%08lx\n", 
-			         (unsigned long)memory);
+			if (egg_secure_warnings)
+				fprintf (stderr, "memory does not belong to gnome-keyring: 0x%08lx\n", 
+				         (unsigned long)memory);
 			ASSERT (0 && "memory does does not belong to gnome-keyring");
 		}
 	}
@@ -1095,6 +1150,19 @@ egg_secure_check (const void *memory)
 	return block == NULL ? 0 : 1;
 } 
 
+void
+egg_secure_validate (void)
+{
+	Block *block = NULL;
+	
+	DO_LOCK ();
+	
+		for (block = all_blocks; block; block = block->next)
+			sec_validate (block);
+		
+	DO_UNLOCK ();
+}
+
 void
 egg_secure_dump_blocks (void)
 {
